/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.aspect.plugin;

/*
 * Adapted from org.apache.maven.plugin.source.JarSourceMojo which was released under the following
 * license:
 * <p>
 * Copyright 2001-2005 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language governing permissions
 * and limitations under the License.
 */
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.UUID;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.apache.maven.archiver.MavenArchiveConfiguration;
import org.apache.maven.archiver.MavenArchiver;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.ArtifactUtils;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.MavenProjectHelper;
import org.codehaus.plexus.archiver.ArchiverException;
import org.codehaus.plexus.archiver.jar.JarArchiver;
import org.codehaus.plexus.util.FileUtils;

/**
 * Aggregates sources from main project, and sources from an optional set of dependencies
 * listed within plugin configuration, into a single archive and attaches to project using
 * 'sources' classifier.
 * 
 * @author <a href="mailto:trygvis@inamo.no">Trygve Laugst&oslash;l</a>
 * @author twolff
 * 
 * @goal attach
 */
public class SourceMojo extends AbstractMojo {

    /**
     * Directory that resources are copied to during the build.
     * 
     * @parameter expression="${project.build.directory}/${project.build.finalName}"
     * @required
     */
    private String workDirectory;

    /**
     * 
     * Project helper injected by maven.
     * 
     * @component
     */
    private MavenProjectHelper projectHelper;

    /**
     * The directory for the generated source attachment.
     * 
     * @parameter expression="${project.build.directory}"
     * @required
     */
    private String outputDirectory;

    /**
     * The name of the source attachment to which '-sources.jar' is appended.
     * 
     * @parameter expression="${project.build.finalName}"
     * @required
     */
    private String finalName;

    /**
     * The maven project.
     * 
     * @parameter expression="${project}"
     * @required
     * @readonly
     * @description "the maven project to use"
     */
    private MavenProject project;

    /**
     * The Jar archiver.
     * 
     * @parameter expression="${component.org.codehaus.plexus.archiver.Archiver#jar}"
     * @required
     */
    private JarArchiver jarArchiver;

    /**
     * List of dependencies from which source files should be collected and aggregated into a
     * single source attachment.
     * 
     * @parameter
     */
    private Dependency[] dependencies;

    /**
     * The maven archiver configuration.
     * 
     * @parameter
     */
    private MavenArchiveConfiguration archive = new MavenArchiveConfiguration();

    private File outputDir;
    private File workDir;
    private File tempDir = new File(System.getProperty("java.io.tmpdir"));

    private static final String[] INCLUDES = new String[] { "**/*.java", "**/*.aj" };

    public void execute() throws MojoExecutionException {

        getLog().info(" ======= SourceMojo settings =======");
        getLog().info("workDirectory[" + workDirectory + "]");
        getLog().info("outputDirectory[" + outputDirectory + "]");
        getLog().info("finalName[" + finalName + "]");
        getLog().info("dependencies[" + dependencies + "]");

        if (dependencies == null) {
            return;
        }

        File sourceAttachment = new File(getOutputDir(), finalName + "-sources.jar");

        MavenArchiver archiver = new MavenArchiver();
        try {
            archiver.setArchiver(jarArchiver);
            archiver.setOutputFile(sourceAttachment);
            jarArchiver.addDirectory(getWorkDir());
            archiver.createArchive(project, archive);
        } catch (Exception ex) {
            throw new MojoExecutionException("Error assembling source attachment. ", ex);
        }

        // add source files from main project

        getLog().info("adding sources from main project");

        Iterator<?> sourceRoots = project.getCompileSourceRoots().iterator();
        while (sourceRoots.hasNext()) {
            File sourceDirectory = new File((String)sourceRoots.next());
            if (sourceDirectory.exists()) {
                try {
                    jarArchiver.addDirectory(sourceDirectory, INCLUDES, FileUtils.getDefaultExcludes());
                } catch (ArchiverException ae) {
                    throw new MojoExecutionException("Error assembling source attachment. ", ae);
                }
            }
        }

        // add source files from listed dependencies

        for (Dependency dependency : dependencies) {
            String groupID = dependency.getGroupId();
            String artifactID = dependency.getArtifactId();
            String key = ArtifactUtils.versionlessKey(groupID, artifactID);
            Artifact artifact = (Artifact)project.getArtifactMap().get(key);
            if (artifact == null) {
                getLog().error("dependency " + key + " not defined on project");
            } else {
                getLog().info("adding sources for dependency " + key);
                if (artifact.getFile() != null) {
                    String jarName = null;
                    if (artifact.isSnapshot()) {
                        jarName = artifactID + "-SNAPSHOT-sources.jar";
                    } else {
                        jarName = artifactID + "-" + artifact.getVersion() + "-sources.jar";
                    }
                    File sourcesJar = new File(artifact.getFile().getParent(), jarName);
                    if (sourcesJar.exists()) {
                        unpackArchive(sourcesJar);
                    } else {
                        getLog().error("sources file " + sourcesJar + " not found");
                    }
                } else {
                    getLog().error("dependency " + key + " has no associated file");
                }
            }
        }

        // create the source archive and attach it to project. it will be
        // installed by install plugin within install phase

        try {
            archiver.createArchive(project, archive);
            projectHelper.attachArtifact(project, "java-source", "sources", sourceAttachment);
        } catch (Exception ex) {
            throw new MojoExecutionException("Error assembling source attachment. ", ex);
        }

    }

    protected File getOutputDir() {
        if (outputDir == null) {
            outputDir = new File(outputDirectory);
            if (!outputDir.exists()) {
                outputDir.mkdir();
            }
        }
        return outputDir;
    }

    protected File getWorkDir() {
        if (workDir == null) {
            workDir = new File(workDirectory);
            if (!workDir.exists()) {
                workDir.mkdir();
            }
        }
        return workDir;
    }

    private void unpackArchive(File archive) throws MojoExecutionException {

        // create a temporary directory to which we will unpack the
        // archive and set delete on exit to clean up after ourself

        File target = new File(tempDir, "source-plugin-" + UUID.randomUUID().toString());
        target.deleteOnExit();
        target.mkdir();

        ZipFile zipFile = null;
        try {
            zipFile = new ZipFile(archive);
            Enumeration<?> entries = zipFile.entries();
            while (entries.hasMoreElements()) {
                ZipEntry entry = (ZipEntry)entries.nextElement();
                File file = new File(target, File.separator + entry.getName());
                if (entry.isDirectory() && !file.exists()) {
                    file.mkdirs();
                } else {
                    InputStream in = null;
                    OutputStream out = null;
                    try {
                        in = zipFile.getInputStream(entry);
                        out = new BufferedOutputStream(new FileOutputStream(file));
                        byte[] buffer = new byte[4096];
                        for (int length = 0; length >= 0; length = in.read(buffer)) {
                            out.write(buffer, 0, length);
                        }
                    } finally {
                        if (in != null) {
                            in.close();
                        }
                        if (out != null) {
                            out.close();
                        }
                    }
                }
            }
        } catch (Exception ex) {
            throw new MojoExecutionException("Error assembling source attachment. ", ex);
        } finally {
            if (zipFile != null) {
                try {
                    zipFile.close();
                } catch (IOException ignore) {
                }
            }
        }

        // add temp directory containing unpacked entries to source attachment
        // archiver

        try {
            jarArchiver.addDirectory(target, INCLUDES, FileUtils.getDefaultExcludes());
        } catch (ArchiverException ae) {
            throw new MojoExecutionException("Error assembling source attachment. ", ae);
        }

    }

}
